
import com.cyberbotics.webots.controller.Robot;
import com.cyberbotics.webots.controller.Servo;
import java.lang.Math.*;

public class MuscleModel {

	private double prevForce1;
	private double prevForce2;
	private double force1;
	private double force2;
	private double totalForce;
	private double spikerateM1;
	private double length;
	private double prevLength;
	private double lengthNormalized;
	private double maxLength;
	private double velocity;
	private Servo parentServo;
	private Robot parentRobot;
	private String name;
	private double prevTime;
	private double minPos;
	private double maxPos;
	private double currentPos;
	private double angleOffset;
	private final double alfaM1;
	private double r;			//lengde pA arm, fra servo til muskelfeste
	private double spikeIncrease;
	private double dT;
	private int muscleDirection;
	private double maxForce;
	private double currentTime;
	private double angleMaxLength;
	private double torque;
	private double fMarked1;
	private double fMarked2;
	private double fa2;
	private double fa3;
	private double fp2;
	private double fp3;
	private double fa11;
	private double fa12;
	private double prevLength2;
	private double angleMinLength;
	private double minLength;
	private double maxLengthChange;
	private double timestep;
	private double maxVelocity;
	private MotorNeuron mNeuron1;
	private MotorNeuron mNeuron2;
	private SensoryNeuron golgiInputNeuron;
	private SensoryNeuron lengthNeuron;
	private SensoryNeuron accNeuron;
	private boolean manualControl;
	private double maxTorque;
	private double spikerateM2;
	private double alfaM2;
//	MyLiveGraph l;
	private double spikeRateAmplifier;
	private double needA;
	private int needB;
	
	public MuscleModel(String name, Servo parentServo, Robot parentRobot, double minPos, double maxPos, int muscleDirection, double r, double maxForce){
		this.name 	= name;
		this.parentServo = parentServo;
		this.parentRobot = parentRobot;
		this.r = r;
		this.minPos		= minPos;
		this.maxPos		= maxPos;
		this.muscleDirection = muscleDirection; //"venstre" muskel = 1, hOyre =-1  /|\
		this.maxTorque= maxForce*r*Math.sin((Math.PI/2 + maxPos)/2);
		this.needA=0;
		this.needB=0;
	
		timestep = parentRobot.getBasicTimeStep();
		alfaM1 = 0.1; //fast twitch
		alfaM2 = 0.9; //slower twitch
		spikeIncrease = 0.1;
		spikeRateAmplifier=7; // define max spike rate as 100Hz or else almost no force is generated due to the low normalized spike rate
		dT = 0; // -> 0 velocity first round
		this.maxForce = maxForce;

		force1		= 0;
		force2		= 0;
		totalForce	= 0;
		prevForce1 	= 0;
		prevForce2 	= 0;
		spikerateM1 = 0;
		spikerateM2 = 0;
		length 		= 0;
		prevLength  = calculateLength();
		velocity 	= 0;
		prevTime 	= 0;
		torque		= 0;
		currentPos=parentServo.getPosition();
		currentTime=parentRobot.getTime();
		angleMaxLength = muscleDirection==1 ? minPos :  maxPos;
		maxLength = calculateLength(angleMaxLength);
		angleMinLength= muscleDirection==1 ? maxPos : minPos;
		minLength = calculateLength(angleMinLength);
		maxLengthChange = maxLength-minLength;
		maxVelocity = maxLengthChange /timestep ; // -> m/tid
		manualControl=false;
		
		//Used to plot muscle forces and spike rates
//		if(name.equals("Muscle L_UP_  _")){
////			l = new MyLiveGraph("Forces") ;
////			l.addDataSeries(new String[]{"Time","Velocity","Length","Torque","Force1","Force2"});
//		}
	}
	public void setMotorNeuron(MotorNeuron mn1, MotorNeuron mn2){
		mNeuron1=mn1;
		mNeuron2=mn2;
	}
	
	public void update() {
		currentTime=parentRobot.getTime();
		currentPos=parentServo.getPosition();
			
//		printVariables();
		
		// Update current time and calculate interval
		if(parentRobot.getTime() != prevTime) dT = currentTime - prevTime;
		prevTime = currentTime;
		
		// Update current length and calculate velocity
		length = calculateLength();
		
		//ADDED TO MAYBE HELP WITH: ODE Message 3: LCP internal error, s <= 0 (s=0.0000e+000)
		if(length<minLength){
			length=minLength;
		}else if(length>maxLength){
			length=maxLength;
		}
		lengthNormalized = (length-minLength)/(maxLength-minLength);
		double deltaLength= length-prevLength;
		
		//   Velocity is relative to 0.5. That is, values greater than 0.5 represent "positive" velocity and values below 0.5 represent "negative" velocity.
		velocity = 0.5+ (( deltaLength / maxLengthChange) / 1)/2;//(dT/(timestep/1000)); // dT/(timestep/1000) = 1  
		

		if(mNeuron1!=null && !manualControl){
			spikerateM1 = spikeRateAmplifier*mNeuron1.getSpikerate(0); 
			spikerateM2 = spikeRateAmplifier*mNeuron2.getSpikerate(0);	
			if(spikerateM1>1){
				spikerateM1=1;
			}
			if(spikerateM2>1){
				spikerateM2=1;
			}
		}
		
		// Update current force
		totalForce = calculateForce(); 
		torque= calculateTorque();	
		
		/* 
		 * UPDATE THE INPUT TO THE ANN
		 */
		if(golgiInputNeuron!=null && !manualControl){		//updates input from layer 8
			golgiInputNeuron.setSensoryValue(3*Math.pow(torque,7)+0.1*torque);
		}
		
		if(lengthNeuron!=null && !manualControl){
			lengthNeuron.setSensoryValue(5*Math.pow(lengthNormalized,7));
			accNeuron.setSensoryValue(5*Math.pow(Math.max(velocity-0.5, 0), 7));	//only positive velocity count if positive  
		}
		
//		if(name.equals("Muscle L_UP_  _")){ //PLotting stuff
//			
////			 l.updateDataSets(new double[]{SpikingNetwork.currentTime,velocity, lengthNormalized, torque, force1, force2});
//		}
		/*
		 * END UPDATE INPUT TO ANN
		 */
		
//		printVariables();

		prevLength = length;
		
		// Spiking rate decay	//used when controlling the agent manually
//		if(spikerateM1 > 1) spikerateM1 = 1;
//		if(spikerateM1 > 0) spikerateM1 -= spikeIncrease/1.5;
//		if(spikerateM1 < 0) spikerateM1 = 0;		
	}
	
	public double calculateLength() {
		double alfaLength = Math.PI/2 - muscleDirection*currentPos;
		return Math.sin(alfaLength/2)*r*2;
	}
	
	public double calculateLength(double position){
		double alfaLength = Math.PI/2 - muscleDirection*position; 
		return Math.sin(alfaLength/2)*r*2;
	}
	
	public double calculateForce(){
		fMarked1=Fmarked(spikerateM1,lengthNormalized, velocity); 
		fMarked2=Fmarked(spikerateM2); // only need to recalculate fa1
		prevForce1=force1;
		prevForce2=force2;
		force1 = alfaM1*prevForce1 + ((1-alfaM1)* fMarked1);
		force2 = alfaM2*prevForce2 + ((1-alfaM2)* fMarked2);
		return force1 + force2;
	}

	public double Fmarked(double spikeRate1, double muscleLength, double muscleVelocity){
		fa11=spikeRate1;
		fa2=Fa2(muscleLength);	//for printing /debug resons
		fa3=Fa3(muscleVelocity);
		fp2=Fp2(muscleLength);
		fp3=Fp3(muscleVelocity);
		return maxForce*(fa11*fa2*fa3 + fp2*fp3);

	}
	public double Fmarked(double spikerate) {
		fa12=spikerate;
		return maxForce*(fa12*fa2*fa3 + fp2*fp3);
	}

	public double Fa2(double muscleLength){
		return -Math.pow((1.4*muscleLength-0.9), 2)+1;
	}
	
	public double Fa3(double muscleVelocity){
		if(muscleVelocity<0.5){
			return Math.pow(Math.exp(muscleVelocity-0.572975),7);
		}else if(muscleVelocity<0.7){
			return -10*Math.pow((muscleVelocity-0.7),2)+1;
		}else{
			return 1;
		}
	}

	public double Fp2(double muscleLength){
		return Math.pow(muscleLength-0.25, 6);
	}
	public double Fp3 (double muscleVelocity){
		return 0.5*muscleVelocity+0.25;
	}

	public double calculateTorque(){
		
		double totAngle= Math.PI/2 - muscleDirection*currentPos;	//   __/   eller  /_
		double alfa= (Math.PI-totAngle)/2;					
		return totalForce*r*Math.cos(alfa);
	}
	
	public double getTorque() {
		return torque;
	}

	public double getSpikerate() {
		return spikerateM1;
	}

	public void updateSpikerate() {
		spikerateM1 += spikeIncrease;
	}
	
	public double getLength() {
		return length;
	}

	public double getVelocity() {
		return velocity;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public double getForce() {
		return totalForce;
	}

	public void setForce(double force) {
		this.totalForce = force;
	}
	
	public double getFp2() {
		return fp2;
	}
	
	public double getFp3() {
		return fp3;
	}
	
	public double getMaxForce() {
		return maxForce;
	}
	
	public void setGolgiNeuron(SensoryNeuron sensoryNeuron) {
		this.golgiInputNeuron = sensoryNeuron;
	}
	
	public void setLengthAccNeurons(SensoryNeuron lengthNeuron, SensoryNeuron accNeuron) {
		this.lengthNeuron = lengthNeuron;
		this.accNeuron = accNeuron;
	}
	
	public void toggleManualControl() {
		manualControl = manualControl==false;
	}
	
	public void setNeedA(double needA) {
		this.needA = needA;
	}
	
	public void setNeedB(int needB) {
		this.needB = needB;
	}
	
	public double getSpikerate2() {
		// TODO Auto-generated method stub
		return spikerateM2;
	}
	
	public void printVariables(){
//		if(name.equals("Muscle fl.1.1")){
//			
////			System.out.println("fMarked:         "+fMarked );
////			System.out.println("Fa1 (spike)      "+fa1);
////			System.out.println("Fa2 (length)     "+fa2);
////			System.out.println("Fa3 (velocity)   "+fa3);
////			System.out.println("Fp2 (length)       "+fp2);
////			System.out.println("Fp3 (velocity)     "+fp3);
////			System.out.println("Normalized length  "+normalizedLength);
////			System.out.println("force:           "+force);
////			System.out.println("prevForce:       "+prevForce);
////double normvelocity = 0.5+((lengthNormalized - (prevLength2/maxLength)) / dT);
////			System.out.println("velocity:        "+velocity);
////			System.out.println("normLength:      "+lengthNormalized);
////			System.out.println("normprevLength:  "+prevLength2);
//////			System.out.println("dT:              "+dT);
////			
////			System.out.println("--------------------------- ");
//			System.out.println("prevLength:      "+prevLength2);
//			System.out.println("MuscleLength:    "+length);
//			System.out.println("MaxChangeLength: "+maxLengthChange);
//			System.out.println("velocity:        "+velocity);
//			System.out.println("Fp2 (length)     "+fp2);
//			System.out.println("Fp3 (velocity)   "+fp3);
//			System.out.println("--------------------------- ");
//		}
		if(name.equals("Muscle fl.1.2")){
		System.out.println("=============================");
		System.out.println("Name:            "+name);
		System.out.println("alfaM:           "+alfaM1);
		System.out.println("spikeIncrease:   "+spikeIncrease);
		System.out.println("maxForce:        "+maxForce);
		System.out.println("maxLength:       "+maxLength);
		System.out.println("angleMaxLength:  "+angleMaxLength);
		System.out.println("muscleDirection: "+muscleDirection);
		
		
		System.out.println("    " );
		System.out.println("force:           "+totalForce);
		System.out.println("prevForce:       "+prevForce1);
		System.out.println("torque:          "+torque);
		System.out.println("position:        "+currentPos);
		System.out.println("MuscleLength:    "+length);
		
		System.out.println("spikerate:       "+spikerateM1);
		System.out.println("prevLength:      "+prevLength);
		System.out.println("velocity:        "+velocity);
		System.out.println("current Time:    "+parentRobot.getTime());
		System.out.println("dT:              "+dT);
		System.out.println("r:               "+r);
		
		System.out.println("--------------------------- ");
		}
	}
}

