import org.LiveGraph.LiveGraph;
import org.LiveGraph.dataFile.write.DataStreamWriter;
import org.LiveGraph.dataFile.write.DataStreamWriterFactory;
import org.LiveGraph.settings.DataFileSettings;

/**
 * This class is used for generating a demo data file for demonstration and
 * testing purposes.
 * It creates a data file and fills it with data. This happens in bursts
 * of 1 to 10 datasets at a time.
 * Between the bursts the execution is paused for 0 to 10 milliseconds.
 * This way a data generating simulation is emulated.
 * The program stops after 100000 datasets have been written.
 * 
 * <p><strong>LiveGraph</strong> (http://www.live-graph.org).</p>
 * <p>Copyright (c) 2007 by G. Paperin.</p>
 * <p>File: LiveGraphDemo.java</p>
 * <p><u>Licence</u>: See http://www.live-graph.org/licence.html.</p>
 *
 * @author Greg Paperin
 * @version 1.0.1
 */
public class MyLiveGraph {

	public static final String DEMO_DIR = System.getProperty("user.dir");
	DataStreamWriter out;

	public MyLiveGraph(String name) {

		// Print a welcome message:
		System.out.println("Welcome to the LiveLog demo.");

		// Setup a data writer object:
//		out = DataStreamWriterFactory.createDataWriter(DEMO_DIR, "SkinnerTest");
		double timeStamp= System.currentTimeMillis();
		timeStamp=44;
		
		out = DataStreamWriterFactory.createDataWriter("LiveGraph/"+name+""+timeStamp+".dat", true);
		
		try{
		DataFileSettings dfs = new DataFileSettings();
		
		dfs.setDataFile("LiveGraph/"+name+""+timeStamp+".dat");
		  dfs.setUpdateFrequency(50);
		  dfs.save("LiveGraph/"+name+""+timeStamp+".lgdfs");  
			
//		  LiveGraph app = LiveGraph.application();
//		  app.exec(new String[] {"-dfs", "LiveGraph/"+name+""+timeStamp+".lgdfs"});
		 
		}catch (Exception e) {
			System.out.println(e);
		}
		
		// Set a values separator:
		out.setSeparator(";");

		// Add a file description line:
		out.writeFileInfo(name);
		
	}
	public void addDataSeries(String[] series){
		
		for(int i= 0;i<series.length;i++){
			out.addDataSeries(series[i]);
		}
	}
	public void updateDataSets(double[] updateData){
//		out.setDataValue(SpikingNetwork.currentTime);
		for(int i=0;i<updateData.length;i++){
			out.setDataValue(updateData[i]);
		}
		
		out.writeDataSet();
	     
	      // Check for IOErrors:     
	      if (out.hadIOException()) {
	        out.getIOException().printStackTrace();
	        out.resetIOException();
	      }
		
	}
	public void endPlot(){
		out.close();
		  System.out.println("Demo finished. Cheers.");
	}
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}

} 