import java.io.*;
import java.util.ArrayList;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;


public class NetworkParser {
	public static int HUME = 0;
	public static int HUME_EXT = 1;
	public static int HUME_INH = 2;
	public static int PAVLOV = 3;
	public static int PAVLOV_EXT = 4;
	public static int PAVLOV_INH = 5;
	public static int SKINNER = 6;
	public static int SKINNER_EXT = 7;
	public static int SKINNER_INH = 8;
	public static int STATIC_SYN = 9;
	public static int RES_NEURON = 1;
	public static int RS_NEURON = 2;
	public static int SENSOR_NEURON = 3;
	public static int MOTOR_NEURON = 4;

	Layer[] layers;
	ArrayList<Synapse> tempSynapses;
	Settings settings;
	String filename;
	Display display;

	BufferedReader fileReader;
	int lineNumber;

	public NetworkParser(String filename, Settings settings, Display display) {
		this.settings = settings;
		this.filename = filename;
		this.display = display;
	}

	public void parseNetwork() {
		String line;
		StringTokenizer tokenizer;
		String token;
		String synapseType;
		int to_id, from_id, to_layer, from_layer;
		tempSynapses = new ArrayList<Synapse>();
		layers = new Layer[10];
		lineNumber = 0;

		try {
			fileReader = new BufferedReader(new FileReader(filename));
			while((line = fileReader.readLine()) != null) {
				lineNumber++;
				tokenizer = new StringTokenizer(line);
				try{
					token = tokenizer.nextToken();
					if (!token.equals("#")) {

						if (token.equals("layer")) {
							int layerID, neuronCount;
							String neuronTypeString;
							String layerName;
							try {
								layerID = Integer.parseInt(tokenizer.nextToken());
								neuronCount = Integer.parseInt(tokenizer.nextToken());
								neuronTypeString = tokenizer.nextToken();
								layerName = tokenizer.nextToken();
								int neuronType;
								if (neuronTypeString.equals("RES")) { neuronType = RES_NEURON; }
								else if (neuronTypeString.equals("RS")) { neuronType = RS_NEURON; }
								else if (neuronTypeString.equals("SENSOR")) { neuronType = SENSOR_NEURON; }
								else if (neuronTypeString.equals("MOTOR")) { neuronType = MOTOR_NEURON; }
								else { throw new Exception("Illegal neuron type: "+neuronTypeString); }

								if (layers[layerID] != null) {
									throw new Exception("Unable to create layer "+layerID+", layer already exists.");
								} else {
									// Create a new layer
									layers[layerID] = new Layer(layerID+" "+layerName, layerID, settings);
									display.setLayer(layers[layerID], layerID);

									// Create a list of neurons
									Neuron[] neurons = new Neuron[neuronCount];
									// Generate n neurons in this layer.
									for(int i=0;i<neuronCount;i++) { 
										if(neuronCount>30){ // RS neurons in the large topology
											if(i>7 && i<10 || i>17 && i<20 || i>27 && i<30 || i>37 && i<40){
												neurons[i] = new RSNeuron(i, settings.getInitActivationLvl(), layers[layerID], settings);
											}else{
												if (neuronType == RES_NEURON) neurons[i] = new RESNeuron(i, settings.getInitActivationLvl(), layers[layerID], settings);
												if (neuronType == RS_NEURON) neurons[i] = new RSNeuron(i, settings.getInitActivationLvl(), layers[layerID], settings);
												if (neuronType == SENSOR_NEURON) neurons[i] = new SensoryNeuron(i, 10, layers[layerID], settings);
												if (neuronType == MOTOR_NEURON) neurons[i] = new MotorNeuron(i, settings.getInitActivationLvl(), layers[layerID], settings);
											}
										}else{
										if (neuronType == RES_NEURON) neurons[i] = new RESNeuron(i, settings.getInitActivationLvl(), layers[layerID], settings);
										if (neuronType == RS_NEURON) neurons[i] = new RSNeuron(i, settings.getInitActivationLvl(), layers[layerID], settings);
										if (neuronType == SENSOR_NEURON) neurons[i] = new SensoryNeuron(i, 10, layers[layerID], settings);
										if (neuronType == MOTOR_NEURON) neurons[i] = new MotorNeuron(i, settings.getInitActivationLvl(), layers[layerID], settings);
										}									
									}
									layers[layerID].addNeurons(neurons);
								}
							} catch (NoSuchElementException e) {
								System.out.println("["+filename+":"+lineNumber+"] Missing one or more arguments to 'layer'. 4 arguments needed; layerID, neuron count and neuron type (RES or RS).");							
							}
						}

						if (token.equals("syn")) {
							try {
								from_layer = Integer.parseInt(tokenizer.nextToken());
								from_id = Integer.parseInt(tokenizer.nextToken());
								synapseType = tokenizer.nextToken();
								to_layer = Integer.parseInt(tokenizer.nextToken());
								to_id = Integer.parseInt(tokenizer.nextToken());
								int delay = Integer.parseInt(tokenizer.nextToken());
								double weight = Double.parseDouble(tokenizer.nextToken());

								int syn_type_int;
								
								if 		(synapseType.equals("H"))  { syn_type_int = HUME; }
								else if (synapseType.equals("H+")) { syn_type_int = HUME_EXT; }
								else if (synapseType.equals("H-")) { syn_type_int = HUME_INH; }
								else if (synapseType.equals("P"))  { syn_type_int = PAVLOV; }
								else if (synapseType.equals("P+")) { syn_type_int = PAVLOV_EXT; }
								else if (synapseType.equals("P-")) { syn_type_int = PAVLOV_INH; }
								else if (synapseType.equals("S"))  { syn_type_int = SKINNER; }
								else if (synapseType.equals("S+")) { syn_type_int = SKINNER_EXT; }
								else if (synapseType.equals("S-")) { syn_type_int = SKINNER_INH; }
								else if (synapseType.equals("ST")) { syn_type_int = STATIC_SYN; }
								else {
									syn_type_int = -1;
									System.out.println("Unable to parse synapse type.");
								}

								createSynapse(from_layer, from_id, syn_type_int, to_layer, to_id, delay, weight);
							} catch (NoSuchElementException e) {
								System.out.println("["+filename+":"+lineNumber+"] Missing one or more arguments to 'syn'. 6 arguments needed: [from_layerID, from_neuronID, synapse_type, to_layerID, to_neuronID, delay].");
								//								throw new Exception("Syntax error: Missing arguments");
							}
						}

						if (token.equals("synall")) {
							try {
								int fromLayerID = Integer.parseInt(tokenizer.nextToken());
								synapseType = tokenizer.nextToken();
								int toLayerID = Integer.parseInt(tokenizer.nextToken());
								int delay = Integer.parseInt(tokenizer.nextToken());
								boolean internal = fromLayerID == toLayerID;
								double weight = Double.parseDouble(tokenizer.nextToken());
								int syn_type_int;

								boolean skiprow=false;
								int rowLength =0;
								if(tokenizer.hasMoreElements()){
									skiprow=true;
									tokenizer.nextToken(); 	//throw away "skipRow" 
									rowLength=Integer.parseInt(tokenizer.nextToken());
								}

								if 		(synapseType.equals("H"))  { syn_type_int = HUME; }
								else if (synapseType.equals("H+")) { syn_type_int = HUME_EXT; }
								else if (synapseType.equals("H-")) { syn_type_int = HUME_INH; }
								else if (synapseType.equals("P"))  { syn_type_int = PAVLOV; }
								else if (synapseType.equals("P+")) { syn_type_int = PAVLOV_EXT; }
								else if (synapseType.equals("P-")) { syn_type_int = PAVLOV_INH; }
								else if (synapseType.equals("S"))  { syn_type_int = SKINNER; }
								else if (synapseType.equals("S+")) { syn_type_int = SKINNER_EXT; }
								else if (synapseType.equals("S-")) { syn_type_int = SKINNER_INH; }
								else if (synapseType.equals("ST")) { syn_type_int = STATIC_SYN; }
								else {
									syn_type_int = -1;
									System.out.println("Unable to parse synapse type.");
								}
								if(skiprow){	//to group motorneurons in layer 3 to not inhibit each other inside one group
									double atRow=0;
									for(int i = 0; i<layers[fromLayerID].getNumberOfNeurons();i++) {
										if(i%rowLength==0 && i!=0){
											atRow++;
										}
										for(int j = 0; j<layers[toLayerID].getNumberOfNeurons();j++) {
											
											if(i!=j){
												if(j>= rowLength*atRow+rowLength || j<rowLength*atRow){
													createSynapse(fromLayerID, i+1, syn_type_int, toLayerID, j+1, delay, weight);
												}else{//create ST synapses inside each motor group
													createSynapse(fromLayerID, i+1, STATIC_SYN, toLayerID, j+1, delay, -weight);	//stronger weight between neurons in motor group
												}
											}
										}
									}			
								}

								else{

									for(int i = 0; i<layers[fromLayerID].getNumberOfNeurons();i++) {
										for(int j = 0; j<layers[toLayerID].getNumberOfNeurons();j++) {
											if (internal){ //avoid making synapse from a neuron to itself
												if(i != j) createSynapse(fromLayerID, i+1, syn_type_int, toLayerID, j+1, delay, weight);
											}else{
												createSynapse(fromLayerID, i+1, syn_type_int, toLayerID, j+1, delay, weight);
											}
										}
									}
								}
							} catch (NoSuchElementException e) {
								System.out.println("["+filename+":"+lineNumber+"] Missing one or more arguments to 'synall'. 4 arguments needed: [fromLayerID, synapse type, toLayerID, delay].");
								//								throw new Exception("Syntax error: Missing arguments");
							}
						}
						if (token.equals("syn_1_to_many")) {
							try {
								int fromLayerID = Integer.parseInt(tokenizer.nextToken());
								synapseType = tokenizer.nextToken();
								int toLayerID = Integer.parseInt(tokenizer.nextToken());
								String seq_mod = tokenizer.nextToken();
								int nodesPrRowColumn = Integer.parseInt(tokenizer.nextToken());
								int neuronsPrMuscle = Integer.parseInt(tokenizer.nextToken());
								int delay = Integer.parseInt(tokenizer.nextToken());
								double weight = Double.parseDouble(tokenizer.nextToken());

								int syn_type_int;

								if 		(synapseType.equals("H"))  { syn_type_int = HUME; }
								else if (synapseType.equals("H+")) { syn_type_int = HUME_EXT; }
								else if (synapseType.equals("H-")) { syn_type_int = HUME_INH; }
								else if (synapseType.equals("P"))  { syn_type_int = PAVLOV; }
								else if (synapseType.equals("P+")) { syn_type_int = PAVLOV_EXT; }
								else if (synapseType.equals("P-")) { syn_type_int = PAVLOV_INH; }
								else if (synapseType.equals("S"))  { syn_type_int = SKINNER; }
								else if (synapseType.equals("S+")) { syn_type_int = SKINNER_EXT; }
								else if (synapseType.equals("S-")) { syn_type_int = SKINNER_INH; }
								else if (synapseType.equals("ST")) { syn_type_int = STATIC_SYN; }
								else {
									syn_type_int = -1;
									System.out.println("Unable to parse synapse type.");
								}

								if(seq_mod.equals("seq")){	//neuronsPrMuscle =1 except in layer 7->4 
									for(int i = 0; i<(layers[fromLayerID].getNumberOfNeurons()-(neuronsPrMuscle-1));i+=neuronsPrMuscle) {
										for(int j = (i/neuronsPrMuscle)*nodesPrRowColumn; j<(i/neuronsPrMuscle)*nodesPrRowColumn+nodesPrRowColumn;j++) {
											
											for(int k=0;k<neuronsPrMuscle;k++){
												createSynapse(fromLayerID, i+1+k, syn_type_int, toLayerID, j+1, delay, weight);
											}
											
										}
									}
								}else if(seq_mod.equals("mod")){
									int counteri=-1;
									for(int i = 0; i<layers[fromLayerID].getNumberOfNeurons()-(neuronsPrMuscle-1);i+=neuronsPrMuscle) {
										counteri++;
										for(int j = 0; j<layers[toLayerID].getNumberOfNeurons();j++) {	
											if((j-counteri)%nodesPrRowColumn ==0){
												for(int k=0;k<neuronsPrMuscle;k++){	
													System.out.println((i+k) +" , "+ j);	
													createSynapse(fromLayerID, i+1+k, syn_type_int, toLayerID, j+1, delay, weight);
												}
												
											}
										}
									}
								}else{
									System.out.println("Unable to parse seq_mod, synapses from Layer " +fromLayerID+ " to Layer "+ toLayerID+ " not created...");
								}
							} catch (NoSuchElementException e) {
								System.out.println("["+filename+":"+lineNumber+"] Missing one or more arguments to 'synall'. 6 arguments needed: [fromLayerID, synapse type, toLayerID, seq_mod, neuronsPrRow, delay].");
								//								throw new Exception("Syntax error: Missing arguments");
							}
						}

						if (token.equals("syn_1to1")) {
							try {
								int from_layerID = Integer.parseInt(tokenizer.nextToken());
								synapseType = tokenizer.nextToken();
								int to_layerID = Integer.parseInt(tokenizer.nextToken());
								int delay = Integer.parseInt(tokenizer.nextToken());
								double weight = Double.parseDouble(tokenizer.nextToken());
								int syn_type_int;
								if 		(synapseType.equals("H"))  { syn_type_int = HUME; }
								else if (synapseType.equals("H+")) { syn_type_int = HUME_EXT; }
								else if (synapseType.equals("H-")) { syn_type_int = HUME_INH; }
								else if (synapseType.equals("P"))  { syn_type_int = PAVLOV; }
								else if (synapseType.equals("P+")) { syn_type_int = PAVLOV_EXT; }
								else if (synapseType.equals("P-")) { syn_type_int = PAVLOV_INH; }
								else if (synapseType.equals("S"))  { syn_type_int = SKINNER; }
								else if (synapseType.equals("S+")) { syn_type_int = SKINNER_EXT; }
								else if (synapseType.equals("S-")) { syn_type_int = SKINNER_INH; }
								else if (synapseType.equals("ST")) { syn_type_int = STATIC_SYN; }
								else {
									syn_type_int = -1;
									System.out.println("Unable to parse synapse type.");
								}

								for(int i = 0; i<layers[from_layerID].getNumberOfNeurons();i++) {
									if (layers[to_layerID].getNodesInLayer().length > i) createSynapse(from_layerID, i+1, syn_type_int, to_layerID, i+1, delay, weight);
								}
							} catch (NoSuchElementException e) {
								System.out.println("["+filename+":"+lineNumber+"] Missing one or more arguments to 'syn_1to1'. 5 arguments needed: [from_layerID, synapse_type, to_layerID, delay, weigth].");
							}
						}

					}


					else { 
						// Current line is a comment (starts with #). Does nothing.
					}
				} catch (NoSuchElementException e) {					
				}
				catch (Exception e) {
					e.printStackTrace(); }
			}
		}
		catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		catch (IOException e) {
			e.printStackTrace(); 
		}
	}

	public void createSynapse(int from_layer, int from_id, int type, int to_layer, int to_id, int delay, double weight) throws Exception {
		try{
			Neuron fromNeuron = layers[from_layer].getNodesInLayer()[from_id-1];
			Neuron toNeuron = layers[to_layer].getNodesInLayer()[to_id-1];
			Synapse synapse;
			
			if 		(type == HUME) 			{ synapse = new HumeSynapse(fromNeuron,toNeuron, 0, 1.0,settings, weight); }
			else if (type == HUME_EXT) 		{ synapse = new HumeSynapse(fromNeuron,toNeuron, 1, 1.0,settings, weight); }
			else if (type == HUME_INH) 		{ synapse = new HumeSynapse(fromNeuron,toNeuron, -1, 1.0,settings,weight); }
			else if (type == PAVLOV) 		{ synapse = new PavlovSynapse(fromNeuron,toNeuron, 0, 1.0,settings, weight); }
			else if (type == PAVLOV_EXT) 	{ synapse = new PavlovSynapse(fromNeuron,toNeuron, 1, 1.0,settings, weight); }
			else if (type == PAVLOV_INH) 	{ synapse = new PavlovSynapse(fromNeuron,toNeuron,-1, 1.0,settings, weight);  }
			else if (type == SKINNER) 		{ synapse = new SkinnerSynapse(fromNeuron,toNeuron, 0, 1.0,settings, weight);  }
			else if (type == SKINNER_EXT) 	{ synapse = new SkinnerSynapse(fromNeuron,toNeuron, 1, 1.0,settings, weight);  }
			else if (type == SKINNER_INH) 	{ synapse = new SkinnerSynapse(fromNeuron,toNeuron, -1, 1.0,settings,weight);  }
			else if (type == STATIC_SYN) 	{
				if(toNeuron instanceof MotorNeuron){
					if(toNeuron.getId()%2 ==0){	//Small motor neuron fire easy -> higher weight
						synapse = new StaticSynapse(fromNeuron,toNeuron,1.0,settings, weight);  
					}else{						//Large motor neuron
						synapse = new StaticSynapse(fromNeuron,toNeuron,1.0,settings, weight);	// /4
					}
				}else{
					synapse = new StaticSynapse(fromNeuron,toNeuron,1.0,settings, weight);
				}
			}
			else { throw new Exception("Illegal Synapse type: "+type); }

			tempSynapses.add(synapse);
			fromNeuron.postSyn.add(synapse);
			synapse.setDelay(delay);
		}
		catch (NullPointerException e) {
			throw new Exception("Could not create synapse, one of the neurons does not exist: ["+from_layer+":"+from_id+", "+to_layer+":"+to_id+"]");
		}

	}

	public Synapse[] getSynapseList() {
		Synapse[] syns = new Synapse[tempSynapses.size()];
		for(int i=0;i<syns.length;i++) {
			syns[i] = tempSynapses.get(i);
		}
		return syns;
	}

	public Layer[] getLayers() {
		return layers;
	}

}
