import java.io.Serializable;
import java.util.ArrayList;


public class Neuron implements Serializable{

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	protected Settings z;
	protected int id;
	protected double activationLvl;
	protected double inputCurrent;
	protected ArrayList<Synapse> postSyn;
	protected int type;
	protected double a;
	protected double b;
	protected double c;
	protected double d;
	protected double v, u;
	protected double initWeight;
	protected int[] postLayerSynId;
	protected Layer parentLayer;
	protected String name;
	protected boolean hasFired;
	protected double Asyn;
	protected ArrayList<Double> fired; //has to be double for visualization
	protected long firingHistory;
	protected int lastFiredTime;		//to calulate alternative spikerate 
	protected int spikeDistance;
	protected double lastTwoSpikeRateConstant;
	protected int spikeRateType;
	protected double spikeRate;
	protected double spikeRateAlfa;
	protected double[] spikeHistory;
	protected int spikeHistoryIndex;
	protected int deltaT;
	protected final int TRACE=0;
	protected final int DISTANCE=1;
	protected final int SIGMOIDWINDOW=2;
	protected final int SPLIT_WINDOW=3;
	
	//skinner
	protected double sumHistoryNew;
	protected double oldestValue;
	protected double sumHistoryOld;
	protected double oldSpikeRateWindow;
	protected double currentSpikeRateWindow;
	protected double currentPostNeuronSpikeRateWindow;
	protected double oldPostNeuronSpikeRateWindow;
	protected double sumHistoryPostNew;
	protected double sumHistoryPostOld;
	//pavlov
	protected double sumHistoryNewP;
	protected double sumHistoryOldP;
	protected double oldSpikeRateWindowP;
	protected double currentSpikeRateWindowP;
	protected double currentPostNeuronSpikeRateWindowP;
	protected double oldPostNeuronSpikeRateWindowP;
	protected double sumHistoryPostNewP;
	protected double sumHistoryPostOldP;
	//hume
	protected double sumHistoryNewH;
	protected double sumHistoryOldH;
	protected double oldSpikeRateWindowH;
	protected double currentSpikeRateWindowH;
	protected double currentPostNeuronSpikeRateWindowH;
	protected double oldPostNeuronSpikeRateWindowH;
	protected double sumHistoryPostNewH;
	protected double sumHistoryPostOldH;
	
	protected int counter;
	protected int windowL1;
	protected int windowL2;
	protected int windowInterval;
	protected int windowL1P;
	protected int windowL2P;
	protected int windowIntervalP;
	protected int windowL1H;
	protected int windowL2H;
	protected int windowIntervalH;
	protected int windowTotalLenght;
	protected int windowTotalLenghtP;
	protected int windowTotalLenghtH;

	public Neuron(int id, double activationLvl, Layer parentLayer, Settings z){
		this.id=id;
		this.z = z;
		this.activationLvl=activationLvl;
		this.parentLayer=parentLayer;
		spikeRateType= z.getSpikeRateType();	//0 = trace, 1= distance , 2 = sigmoid window, 3 = splitWindow
		spikeRate=0;
		spikeRateAlfa= z.getSpikeRateAlfa();
		deltaT = z.getDeltaT();
		counter =0;
		if(spikeRateType==SPLIT_WINDOW){
			windowL1=z.getSRWindowL1()/2;	//split the windows
			windowL2=z.getSRWindowL2()/2;
			windowL1P=z.getSRWindowL1P()/2;
			windowL2P=z.getSRWindowL2P()/2;
			windowL1H=z.getSRWindowL1H()/2;
			windowL2H=z.getSRWindowL2H()/2;
			
			windowInterval =z.getSRWindowInterval();
			windowTotalLenght= windowL1+windowL2+windowInterval;
			
			windowIntervalP =z.getSRWindowIntervalP();
			windowTotalLenghtP= windowL1P+windowL2P+windowIntervalP;
			
			windowIntervalH =z.getSRWindowIntervalH();
			windowTotalLenghtH= windowL1H+windowL2H+windowIntervalH;
			
			int windowMax = Math.max(windowTotalLenghtH, Math.max(windowTotalLenght, windowTotalLenghtP));
			spikeHistory=new double[windowMax];
//			spikeHistory=new double[2*deltaT+offsetSpikeRate];
// <        L1		   ]              [      L2 	   >			          
// [ . . . . [ . . . . .Interval . . . . . . .] . . . . ]
//													   ^													
//												SpikingNetwork.currentTime()
			sumHistoryNew=0;
			sumHistoryOld=0;
			oldestValue=0;
			sumHistoryPostNew=0;
			sumHistoryPostOld=0;
			currentPostNeuronSpikeRateWindow=0;
			oldPostNeuronSpikeRateWindow=0;

			sumHistoryNewP=0;
			sumHistoryOldP=0;
			sumHistoryPostNewP=0;
			sumHistoryPostOldP=0;
			currentPostNeuronSpikeRateWindowP=0;
			oldPostNeuronSpikeRateWindowP=0;
			
			sumHistoryNewH=0;
			sumHistoryOldH=0;
			sumHistoryPostNewH=0;
			sumHistoryPostOldH=0;
			currentPostNeuronSpikeRateWindowH=0;
			oldPostNeuronSpikeRateWindowH=0;
			
			spikeHistoryIndex=-1;
		}
		a=Globals.A;
		b=Globals.B;
		c=Globals.C;
		d=Globals.D;
		v = c; // Initialize v to default lvl.
		u = 0;
		
		oldSpikeRateWindow=0;
		currentSpikeRateWindow=0;
		lastFiredTime = 0;
		spikeDistance = 1000;
		name="Neuron "+parentLayer.getLayerId()+"."+id;
		lastTwoSpikeRateConstant= z.getLastTwoSpikerateConstant();
		postSyn=new ArrayList<Synapse>();
		firingHistory=0;
		fired=new ArrayList<Double>();//[Globals.timesteps];
		inputCurrent=0;
	}

	@SuppressWarnings("static-access")
	public void checkFire() {
		/* Main function method of the neuron, checks whether the neuron shall fire */
		hasFired = false;

		if(Math.random()<z.getRandomBursting()){
			counter=1000;	//spike
		}
		else{
			// Calculated in two steps to avoid "overshooting", i.e. that a value increases too rapidly. // this can lead to v changing from - to + when inpcur >>0
			v += (0.04*Math.pow(v, 2) + 5 * v - u + 140 + inputCurrent)/2;  
			v += (0.04*Math.pow(v, 2) + 5 * v - u + 140 + inputCurrent)/2;
			u += a*( b * v - u ); 
		}
		// Shifting the history queue by one, adding a zero to the end of it.
		firingHistory = firingHistory << 1;
		spikeRate= spikeRateAlfa*spikeRate ;	//slowly decrease the trace based spikerate

		if(spikeRateType==SPLIT_WINDOW){
			spikeHistoryIndex++;
			if(spikeHistoryIndex==spikeHistory.length){
				spikeHistoryIndex=0;
			}
			oldestValue=spikeHistory[spikeHistoryIndex];	//store oldes value before it is replaced
			spikeHistory[spikeHistoryIndex]=0;				//changed to 1 if fire
		}
		if(v>activationLvl || counter>0 && counter%10==0) {
			counter--;//random burst
			if(spikeRateType==SPLIT_WINDOW){
				spikeHistory[spikeHistoryIndex]=1;
			}
			spikeRate+= (1-spikeRateAlfa);		//increase the spikerate a little if spike //type=TRACE 
			
			v = c;
			u += d;
			
			for(int i=0; i<postSyn.size();i++) {
				postSyn.get(i).setPreNodeFired();
			}
			// Adding a 1 to the end (least significant) of the queue.
			firingHistory += 1; 
			hasFired = true;
			parentLayer.addFiredToPlot(this);
			spikeDistance = SpikingNetwork.currentTime-lastFiredTime;
			lastFiredTime = SpikingNetwork.currentTime;
		}
		// Reset the inputlevel before a new iteration.
		inputCurrent = 0;
		if(spikeRateType==SPLIT_WINDOW){
			updateCurrentSpikeRateWindow();
			updateOldSpikeRateWindow();
			updateCurrentPostNeuronSpikeRate();
			updateOldPostNeuronSpikeRate();
		}
	}

	public void updateCurrentSpikeRateWindow(){
//		[---------|currentOld]-----[-------|------]
		int index=spikeHistoryIndex;
		int indexP=spikeHistoryIndex;
		int indexH=spikeHistoryIndex;
		index=index-windowInterval-windowL2;
		indexP=indexP-windowIntervalP-windowL2P;
		indexH=indexH-windowIntervalH-windowL2H;
		if(index<0){
			index=spikeHistory.length+index;
		}
		if(indexP<0){
			indexP=spikeHistory.length+indexP;
		}
		if(indexH<0){
			indexH=spikeHistory.length+indexH;
		}
		sumHistoryNew-=spikeHistory[index];		//remove oldest
		sumHistoryNewP-=spikeHistory[indexP];	//remove oldest
		sumHistoryNewH-=spikeHistory[indexH];	//remove oldest

		index+=windowL1;
		indexP+=windowL1P;
		indexH+=windowL1H;
		if(index>=spikeHistory.length){
			index-=spikeHistory.length;
		}
		if(indexP>=spikeHistory.length){
			indexP-=spikeHistory.length;
		}
		if(indexH>=spikeHistory.length){
			indexH-=spikeHistory.length;
		}
		sumHistoryNew+= spikeHistory[index];	//add newest
		sumHistoryNewP+= spikeHistory[indexP];	//add
		sumHistoryNewH+= spikeHistory[indexH];	//newes

		currentSpikeRateWindow = sumHistoryNew/(double)windowL1;
		currentSpikeRateWindowP = sumHistoryNewP/(double)windowL1P;
		currentSpikeRateWindowH = sumHistoryNewH/(double)windowL1H;

	}
	public void updateOldSpikeRateWindow(){
//		[ Oldold |--------]-----[-------|------]
		int index=spikeHistoryIndex;
		int indexP=spikeHistoryIndex;
		int indexH=spikeHistoryIndex;
		index=index-windowInterval-windowL2;
		indexP=indexP-windowIntervalP-windowL2P;
		indexH=indexH-windowIntervalH-windowL2H;
		if(index<0){
			index=spikeHistory.length+index;
		}
		if(indexP<0){
			indexP=spikeHistory.length+indexP;
		}
		if(indexH<0){
			indexH=spikeHistory.length+indexH;
		}
		sumHistoryOld+=spikeHistory[index];	//add newest
		sumHistoryOldP+=spikeHistory[indexP];	//add newest
		sumHistoryOldH+=spikeHistory[indexH];	//add newest

		if(windowTotalLenght==spikeHistory.length){
			sumHistoryOld-= oldestValue;	//subtract oldest the one just kicked out of the queue	
		}else{
			index=spikeHistoryIndex-windowTotalLenght;
			if(index<0){
				index=spikeHistory.length+index;
			}
			sumHistoryOld-= spikeHistory[index];
		}
		if(windowTotalLenghtP==spikeHistory.length){
			sumHistoryOldP-= oldestValue;	//subtract oldest the one just kicked out of the queue	
		}else{
			indexP=spikeHistoryIndex-windowTotalLenghtP;
			if(indexP<0){
				indexP=spikeHistory.length+indexP;
			}
			sumHistoryOldP-= spikeHistory[indexP];
		}
		if(windowTotalLenghtH==spikeHistory.length){
			sumHistoryOldH-= oldestValue;	//subtract oldest the one just kicked out of the queue	
		}else{
			indexH=spikeHistoryIndex-windowTotalLenghtH;
			if(indexH<0){
				indexH=spikeHistory.length+indexH;
			}
			sumHistoryOldH-= spikeHistory[indexH];
		}
		
		oldSpikeRateWindow= sumHistoryOld/(double)windowL1;
		oldSpikeRateWindowP= sumHistoryOldP/(double)windowL1P;
		oldSpikeRateWindowH= sumHistoryOldH/(double)windowL1H;

	}
	public void updateCurrentPostNeuronSpikeRate(){
//		[ --------|--------]-----[-------|CurrenPo]
		int index=spikeHistoryIndex;
		int indexP=spikeHistoryIndex;
		int indexH=spikeHistoryIndex;
		sumHistoryPostNew+=spikeHistory[index];	//add newest
		sumHistoryPostNewP+=spikeHistory[indexP];	//add newest
		sumHistoryPostNewH+=spikeHistory[indexH];	//add newest

		index-=windowL2;
		indexP-=windowL2P;
		indexH-=windowL2H;

		if(index<0){
			index=spikeHistory.length+index;
		}
		if(indexP<0){
			indexP=spikeHistory.length+indexP;
		}
		if(indexH<0){
			indexH=spikeHistory.length+indexH;
		}
		sumHistoryPostNew-= spikeHistory[index];	//subtract oldest
		sumHistoryPostNewP-= spikeHistory[indexP];	//subtract oldest
		sumHistoryPostNewH-= spikeHistory[indexH];	//subtract oldest

		currentPostNeuronSpikeRateWindow = sumHistoryPostNew/(double)windowL2;
		currentPostNeuronSpikeRateWindowP = sumHistoryPostNewP/(double)windowL2P;
		currentPostNeuronSpikeRateWindowH = sumHistoryPostNewH/(double)windowL2H;
	}
	public void updateOldPostNeuronSpikeRate(){ 
//		[ --------|--------]-----[newPold|------]
		int index=spikeHistoryIndex;
		int indexP=spikeHistoryIndex;
		int indexH=spikeHistoryIndex;
		index-=windowL2;
		indexP-=windowL2P;
		indexH-=windowL2H;
		if(index<0){
			index=spikeHistory.length+index;
		}
		if(indexP<0){
			indexP=spikeHistory.length+indexP;
		}
		if(indexH<0){
			indexH=spikeHistory.length+indexH;
		}
		sumHistoryPostOld+=spikeHistory[index];	//add newest
		sumHistoryPostOldP+=spikeHistory[indexP];	//add newest
		sumHistoryPostOldH+=spikeHistory[indexH];	//add newest

		index-=windowL2;
		indexP-=windowL2P;
		indexH-=windowL2H;

		if(index<0){
			index=spikeHistory.length+index;
		}
		if(indexP<0){
			indexP=spikeHistory.length+indexP;
		}
		if(indexH<0){
			indexH=spikeHistory.length+indexH;
		}
		sumHistoryPostOld-= spikeHistory[index];	//subtract oldest
		sumHistoryPostOldP-= spikeHistory[indexP];	//subtract oldest
		sumHistoryPostOldH-= spikeHistory[indexH];	//subtract oldest

		oldPostNeuronSpikeRateWindow = sumHistoryPostOld/(double)windowL2;
		oldPostNeuronSpikeRateWindowP = sumHistoryPostOldP/(double)windowL2P;
		oldPostNeuronSpikeRateWindowH = sumHistoryPostOldH/(double)windowL2H;
	}
	public double getCurrentRealSpikeRateWindow() {
		return currentPostNeuronSpikeRateWindow;
	}
	public double getOldRealSpikeRateWindow() {
		return oldPostNeuronSpikeRateWindow;
	}

	public double getOldSpikeRateWindow() {
		return oldSpikeRateWindow;
	}
	public double getCurrentSpikeRateWindow() {
		return currentSpikeRateWindow;
	}
	public double getCurrentRealSpikeRateWindowP() {
		return currentPostNeuronSpikeRateWindowP;
	}
	public double getOldRealSpikeRateWindowP() {
		return oldPostNeuronSpikeRateWindowP;
	}
	public double getOldSpikeRateWindowP() {
		return oldSpikeRateWindowP;
	}
	public double getCurrentSpikeRateWindowP() {
		return currentSpikeRateWindowP;
	}
	public double getCurrentRealSpikeRateWindowH() {
		return currentPostNeuronSpikeRateWindowH;
	}
	public double getOldRealSpikeRateWindowH() {
		return oldPostNeuronSpikeRateWindowH;
	}
	
	public double getOldSpikeRateWindowH() {
		return oldSpikeRateWindowH;
	}
	public double getCurrentSpikeRateWindowH() {
		return currentSpikeRateWindowH;
	}

	public double getSpikerate(int delay){
		if(spikeRateType==TRACE){	//trace
			return spikeRate;
		}
		else if(spikeRateType==DISTANCE){	//distance
			// NOT IN USE 
			/*
			double sumHistory= 0;
			for(int i=spikeHistoryIndex;i>spikeHistoryIndex-deltaT;i--){	
				if(i<0){
					sumHistory+=spikeHistory[spikeHistory.length+i];
				}else{
					sumHistory+=spikeHistory[i];
				}
			}
			//		spikeRate=(spikeRate +(sumHistory/spikeRateHistory.length))/2;	//last spike count more
			spikeRate=(sumHistory)/(deltaT);	//last spike count more
			 
			return spikeRate;//lastTwoSpikeRateConstant*1/(spikeDistance+(Math.max(SpikingNetwork.currentTime-lastFiredTime-spikeDistance,0)));
			 */
			return -1;
		}
		else if(spikeRateType==SIGMOIDWINDOW){	//sigmoid window 
			double sigmoidValue=0;
			for(int i=delay; i< Long.SIZE; i++) {
				if ((firingHistory & (1L << i)) != 0)		//if binary value at pos i from left to right is 1
				{
					sigmoidValue += sigmoid(Long.SIZE-i);	//increase sigmoid value by 64-i since we want the last spike to count the most
				}
			}
			return sigmoidValue/Globals.maxSigmoidSpikeRate;		//normalize
		}
		else if(spikeRateType==SPLIT_WINDOW){
			return currentPostNeuronSpikeRateWindow; 
		}
		else{
			return -1;	
		}

	}

	public static double sigmoid(double x){
		return 1/(1+Math.exp(-x/6 +5));		//Edited to fit history of 64 spikes
	}

	public void increaseInput(double psc, String preNodeName) {
		this.inputCurrent += psc;
	}

	public void fireToSynapses() {
		for(int i=0; i<postSyn.size(); i++) {
			postSyn.get(i).setPreNodeFired();
		}
	}
	public ArrayList<Synapse> getPostSynapses(){
		return postSyn;
	}
	public double getV() {
		return v;
	}
	public void setV(double v) {
		this.v = v;
	}
	public double getU() {
		return u;
	}
	public String getName(){
		return name;
	}
	public boolean isHasFired() {
		return hasFired;
	}
	public void setHasFired(boolean hasFired) {
		this.hasFired = hasFired;
	}
	public double[] getFired(){
		double[] res=new double[fired.size()];

		for(int i=0;i<res.length;i++){
			res[i]=fired.get(i);
		}
		return res;
	}
	public int getType(){
		return type;
	}
	public int getId(){
		return id;
	}
	public double getInputCurrent() {
		return inputCurrent;
	}
}
