import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Stroke;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import java.util.ArrayList;

import javax.swing.JFrame;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.DefaultXYDataset;
import org.jfree.util.ShapeUtilities;

public class SpikingNetwork implements Runnable {
	public Layer[] layers;
	public Synapse[] synapses;
	public int timeSteps;
	public int numberOfInputLayers;
	public static int currentTime;
	private Settings z;
	protected GUI gui;

	//CHART COLORS
	private Color gridColor=Color.gray;
	private Color chartBGColor= Color.WHITE;
	private Color colorRS= new Color(0, 210, 0);
	private Color colorRES = new Color(0,0,200);
	private Color colorINH = new Color(200, 0, 0);
	private Color color_AVG = new Color(0, 180, 160);
	private Color colorINP= new Color(50, 50, 50);
	private Stroke rsStroke = new BasicStroke(1.5f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 1.0f, new float[] { 9.0f, 3.0f, 1.0f, 3.0f}, 0.0f);
	private Stroke resStroke = new BasicStroke(1.5f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 1.0f, new float[] { 3.0f }, 0.0f);
	private Stroke chartStroke = new BasicStroke(1.5f);

	private boolean lastRun;
	private String topologyFile;
	private Display display;
	private MuscleModel[] muscles;
	private MyForceC controller;

	public SpikingNetwork(Settings z, GUI gui, MyForceC controller){
		this.controller= controller;
		this.gui=gui;
		this.timeSteps=z.getTimesteps();
		this.z = z;
		topologyFile=z.getTopologyFile();
	}

	public void run() {
		// Looping several times if so is chosen.
		for(int runs=0;runs<1;runs++) { 	//z.getNumberOfRuns() removed posibility for several runs
			Globals.runNumber++;
			display = gui.getDisplay();
			if(topologyFile.endsWith(".dat")){	//loading old network
				try {
					loadTopology(topologyFile);
				} catch (IOException e) {
					e.printStackTrace();
				} catch (ClassNotFoundException e) {
					e.printStackTrace();
				}
				for(int i=0;i<layers.length;i++){
					if(layers[i]!=null){
						display.setLayer(layers[i], i);
					}
				}
				System.out.println("Intital synapse weights: ");
				gui.printSynapses();
				System.out.println("");
			}else{								//parse network from file
				NetworkParser parser = new NetworkParser(topologyFile, z, display);
				parser.parseNetwork();
				layers = parser.getLayers();
				synapses = parser.getSynapseList();

				for(int i= 0; i<synapses.length;i++){
					Synapse syn=synapses[i];
					if(z.getWeightFactor()>0 && !syn.getType().equals("ST")){
						int sign= syn.getCurrentWeight()>0 ? 1:-1;
						syn.setWeight(sign*Math.random()*z.getWeightFactor());	
					}
					if(syn.preNode.parentLayer==syn.postNode.parentLayer){	//if internal layer synapses
						if(z.getInternalDelay()!=-1){
							syn.setDelay(z.getInternalDelay());				//SET DELAY WITHIN LAYERS
						}
						if(syn.preNode.parentLayer.layerId==4){				//Set Asyn Layer 4
							synapses[i].setAsyn(z.getAsyn4());
						}
						else if(syn.preNode.parentLayer.layerId==3){		//Set Asyn Layer 3
							synapses[i].setAsyn(z.getAsyn3());
						}
					}else{
						if(z.getInterLayerDelay()!=-1){
							syn.setDelay(z.getInterLayerDelay());			//SET DELAY BETWEEN LAYERS
						}
						if(syn.preNode.parentLayer.layerId==4 && syn.postNode.parentLayer.layerId==3){
							synapses[i].setAsyn(z.getAsyn4_3());
						}
						else if(syn.preNode.parentLayer.layerId==3 && syn.postNode.parentLayer.layerId==5){			//Set Asyn layer 3 -> Layer 5
							synapses[i].setAsyn(z.getAsyn3_5());
						}
						else if(syn.preNode.parentLayer.layerId==3 && syn.postNode.parentLayer.layerId==6){	//Set Asyn layer 3 -> Layer 6
							synapses[i].setAsyn(z.getAsyn3_6());
						}
						else if(syn.postNode.parentLayer.layerId==1){		//Set Asyn layer 3 -> Layer 1 (Motor)
							synapses[i].setAsyn(z.getAsyn3_1());
						}
						else if(syn.preNode.parentLayer.layerId==7 ){		//Set Asyn layer 7 -> Layer 4/3
							synapses[i].setAsyn(z.getAsyn7_4());
						}
						else if(syn.preNode.parentLayer.layerId==8 && syn.postNode.parentLayer.layerId==5){		//Set Asyn layer 8 -> Layer 5
							synapses[i].setAsyn(z.getAsyn8_5());
						}
						else if(syn.preNode.parentLayer.layerId==9 && syn.postNode.parentLayer.layerId==6){		//Set Asyn layer 9 -> Layer 6
							synapses[i].setAsyn(z.getAsyn9_6());
						}
						else if(syn.preNode.parentLayer.layerId==5 && syn.postNode.parentLayer.layerId==3){		//Set Asyn layer 5 -> Layer 3
							synapses[i].setAsyn(z.getAsyn5_3());
						}
						else if(syn.preNode.parentLayer.layerId==5 && syn.postNode.parentLayer.layerId==4){		//Set Asyn layer 5 -> Layer 4
							synapses[i].setAsyn(z.getAsyn5_4());
						}
						else if(syn.preNode.parentLayer.layerId==6 && syn.postNode.parentLayer.layerId==3){		//Set Asyn layer 6 -> Layer 3
							synapses[i].setAsyn(z.getAsyn6_3());
						}
						else if(syn.preNode.parentLayer.layerId==6 && syn.postNode.parentLayer.layerId==4){		//Set Asyn layer 6 -> Layer 4
							synapses[i].setAsyn(z.getAsyn6_4());
						}
					}
				}
				System.out.println("Intital synapse weights: ");
				gui.printSynapses();
				System.out.println("");
			}

			if(controller!=null){	//NULL WHEN TESTING WITHOUT WEBOTS
				muscles = controller.getMuscleModels();		
				Neuron[] motorNeurons =  layers[1].getNodesInLayer();	//Motor neurons

				int k=0;
				for(int i=0;i<motorNeurons.length/2; i++){ 
					muscles[i].setMotorNeuron((MotorNeuron)motorNeurons[k], (MotorNeuron)motorNeurons[k+1]);
					k+=2;
				}

				Neuron[] golgiInput = layers[8].getNodesInLayer();		//oppdateres fra musclemodel
				Neuron[] lengthAccInput = layers[7].getNodesInLayer();	//oppdateres fra musclemodel
				Neuron[] needInput = layers[9].getNodesInLayer();		//oppdateres fra controller 

				for(int i=0;i<golgiInput.length; i++){ 
					muscles[i].setGolgiNeuron((SensoryNeuron)golgiInput[i]);
				}

				int j=0;

				if(lengthAccInput!=null){
					for(int i=0;i<lengthAccInput.length/2;i++){
						muscles[i].setLengthAccNeurons((SensoryNeuron)lengthAccInput[j],(SensoryNeuron)lengthAccInput[j+1]);
						j+=2;
					}
				}
				controller.setGoalPos(new double[]{z.getGoalPosX(),z.getGoalPosY(),z.getGoalPosZ()});
				controller.setNeedNeurons((SensoryNeuron)needInput[0],(SensoryNeuron)needInput[1]);
				currentTime=1;
				MyForceC.controllerTime=currentTime;	//start off at same time
			}

			Globals.runInProgress=true;
			currentTime=1;
			
			while(currentTime<timeSteps){

				// Update all nodes in all layers:
				for(int i=0;i<layers.length;i++){
					if(layers[i] != null) layers[i].updateNodes();
				}

				// Update all synapses:
				for(int i=0;i<synapses.length;i++){
					synapses[i].updateSpikes();
				}
				
				// Update weights:
				if((currentTime)%z.getWeightUpdateRate() == 0) {
					for(int i=0;i<synapses.length;i++){
						synapses[i].updateWeight();
					}
				}

				//update display
				if(currentTime%(int)(1000/z.getDefaultFPS()) == 0) {
					display.update();
				}
				
				//update progress bar
				gui.getProgressBar().setValue(currentTime + z.getTimesteps()*(Globals.runNumber-1));
				
				currentTime++;
			}
			
			System.out.println("Simulation Complete!");
			displayFireings(false);
		}
	}

	public Synapse[] getSynapses() {
		return synapses;
	}
	private void loadTopology(String url)throws IOException, ClassNotFoundException{
		FileInputStream fis = new FileInputStream(url);
		ObjectInputStream ois = new ObjectInputStream(fis);
		layers =  (Layer[]) ois.readObject();
		synapses = (Synapse[]) ois.readObject();
		ois.close();
	}
	public Layer[] getLayers(){
		return layers;
	}
	
	//STATISTICS  PLOTTING STUFF
	public void displayFireings(boolean showFiringsChart){
		//		plotting single Spike:
		//		DefaultXYDataset dataMembraneP = new DefaultXYDataset();
		//		double []time=new double[5000];
		//		for(int i=0; i<5000;i++){
		//			time[i]=i;
		//		}
		//		
		//		XYLineAndShapeRenderer rendForce=new XYLineAndShapeRenderer(true, false);
		//		
		//		
		//		dataMembraneP.addSeries("Membrane potential", new double[][]{time,Globals.MembranePotential});
		//		rendForce.setSeriesPaint(dataMembraneP.getSeriesCount()-1, Color.BLACK);
		//		rendForce.setSeriesStroke(dataMembraneP.getSeriesCount()-1, new BasicStroke(1.5f));
		//
		//		JFreeChart weightChart  = ChartFactory.createXYLineChart(
		//				"Membrane potential of a RES neuron spiking","Time ms",
		//				"Membrane Potential", dataMembraneP,PlotOrientation.VERTICAL,true, true, false);
		//		weightChart.getPlot().setBackgroundPaint(Color.WHITE);//new Color(0xBFEFFF));
		//		weightChart.setBackgroundPaint(Color.WHITE);
		//		weightChart.getXYPlot().setRenderer(rendForce);
		//		weightChart.getXYPlot().setDomainGridlinesVisible(true);  
		//		weightChart.getXYPlot().setRangeGridlinesVisible(true);  
		//		weightChart.getXYPlot().setRangeGridlinePaint(Color.black);  
		//		weightChart.getXYPlot().setDomainGridlinePaint(Color.black);
		//		weightChart.getLegend().setPosition(RectangleEdge.BOTTOM);
		//		JFrame forceframe = new JFrame();
		//		forceframe.setContentPane(new ChartPanel(weightChart));
		//		forceframe.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		//		forceframe.setLocation(772, 0);
		//		forceframe.pack();
		//		forceframe.setVisible(true);

		for(int i=0; i<layers.length;i++){
			if(layers[i]!=null){

				//ALL FIRINGS =====================================================================

				ArrayList<int[]> allFirRs=layers[i].getAllRsFirings(); // allFirings contain: id and currentTime
				ArrayList<int[]> allFirIn=layers[i].getAllInFirings();
				ArrayList<int[]> allFirRes=layers[i].getAllResFirings();

				if(z.isShowFiringChart()|| showFiringsChart) {
					DefaultXYDataset dataAllFirings = new DefaultXYDataset();

					boolean displayAllFirings=z.getDisplayAverageGraph(); //AVOID DRAWING AVERAGE IF EITHER RS OR RES DOES NOT EXSIT

					XYLineAndShapeRenderer rendAllFirings=new XYLineAndShapeRenderer(false, true);
					rendAllFirings.setSeriesShape(0, ShapeUtilities.createDiamond(1));
					rendAllFirings.setSeriesShape(1, ShapeUtilities.createDiamond(1));
					rendAllFirings.setSeriesShape(2, ShapeUtilities.createDiamond(1));
					rendAllFirings.setSeriesShape(3, ShapeUtilities.createDiamond(1));

					//RS neurons
					double[] xRs = null;
					double[] yRs = null;
					if(allFirRs.size()>0){	
						xRs=new double[allFirRs.size()];
						yRs=new double[allFirRs.size()];

						for(int j=0;j<allFirRs.size();j++){
							xRs[j]=allFirRs.get(j)[1];
							yRs[j]=allFirRs.get(j)[0];

						}
						dataAllFirings.addSeries("Rs Neurons",new double[][]{xRs,yRs});
						rendAllFirings.setSeriesPaint(dataAllFirings.getSeriesCount()-1, colorRS);
					}
					else{
						displayAllFirings=false;
					}

					//Res neurons
					double[] xRes = null;
					double[] yRes = null;
					if(allFirRes.size()>0){
						xRes=new double[allFirRes.size()];
						yRes=new double[allFirRes.size()];

						for(int j=0;j<allFirRes.size();j++){
							xRes[j]=allFirRes.get(j)[1];
							yRes[j]=allFirRes.get(j)[0];
						}
						dataAllFirings.addSeries("Res Neurons",new double[][]{xRes,yRes});
						rendAllFirings.setSeriesPaint(dataAllFirings.getSeriesCount()-1, colorRES);
					}else{
						displayAllFirings=false;
					}

					//inhibitory neurons
					double[] xIn = null;
					double[] yIn = null;
					if(allFirIn.size()>0){
						xIn=new double[allFirIn.size()];
						yIn=new double[allFirIn.size()];

						for(int j=0;j<allFirIn.size();j++){
							xIn[j]=allFirIn.get(j)[1];
							yIn[j]=allFirIn.get(j)[0];
						}
						dataAllFirings.addSeries("Inhibitory Neurons",new double[][]{xIn,yIn});
						rendAllFirings.setSeriesPaint(dataAllFirings.getSeriesCount()-1, colorINH);
					}


					JFreeChart chart = ChartFactory.createXYLineChart(
							"Firings: "+layers[i].getLayerName(),"Time ms","Node nr",
							dataAllFirings,PlotOrientation.VERTICAL,true, true, false);
					chart.getPlot().setBackgroundPaint(chartBGColor);//new Color(0xBFEFFF));
					chart.setBackgroundPaint(Color.WHITE);
					chart.getXYPlot().setRenderer(rendAllFirings);
					chart.getXYPlot().setDomainGridlinesVisible(true);  
					chart.getXYPlot().setRangeGridlinesVisible(true);  
					chart.getXYPlot().setRangeGridlinePaint(Color.black);  
					chart.getXYPlot().setDomainGridlinePaint(Color.black);

					JFrame frame = new JFrame("Spiking network");
					frame.setContentPane(new ChartPanel(chart));
					frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
					frame.setLocation(272, 0);
					frame.pack();
					frame.setVisible(true);
					this.gui.getWindowFrames().add(frame);
				}

				//FOLLOWING (COMMENTED) CODE IS FOR DISPLAYING SINGLE RS NEURON v and u 
				//				if(layers[1].getNodesInLayer().length<2){
				//					ArrayList<double[]> rec = Globals.rsNeuronRecovery;
				//					ArrayList<double[]> acv = Globals.rsNeuronActivation;
				//
				//					System.out.println(rec.size());
				//
				//					DefaultXYDataset data5 = new DefaultXYDataset();
				//
				//					double[] xr=new double[timeSteps];
				//					double[] yr=new double[timeSteps];
				//					double[] xa=new double[timeSteps];
				//					double[] ya=new double[timeSteps];
				//
				//					for(int q = 0; q<rec.size();q++){
				//						xr[q]=rec.get(q)[0];
				//						yr[q]=rec.get(q)[1];
				//
				//						xa[q]=acv.get(q)[0];
				//						ya[q]=acv.get(q)[1];
				//					}
				//					data5.addSeries("RS rec", new double[][]{xr,yr});
				//					data5.addSeries("RS acv", new double[][]{xa,ya});
				//					JFreeChart chart5 = ChartFactory.createXYLineChart(
				//							"Firings: "+layers[i].getLayerName(),"Time ms","u",
				//							data5,PlotOrientation.VERTICAL,true, true, false);
				//					chart5.getPlot().setBackgroundPaint(Color.WHITE);//new Color(0xBFEFFF));
				//					chart5.setBackgroundPaint(new Color(0x5CACEE));
				//					XYLineAndShapeRenderer rend5=new XYLineAndShapeRenderer(true, false);
				//
				//
				//					rend5.setSeriesPaint(0, colorRS);
				//					rend5.setSeriesPaint(1, colorRES);
				//
				//					chart5.getXYPlot().setRenderer(rend5);
				//
				//					JFrame frame5 = new JFrame("Spiking network");
				//					frame5.setContentPane(new ChartPanel(chart5));
				//					frame5.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
				//					frame5.setLocation(272, 0);
				//					frame5.pack();
				//					frame5.setVisible(true);
				//					this.gui.getWindowFrames().add(frame5);
				//				} 

			}
		}
	}
}
