import java.awt.BasicStroke;
import java.awt.Color;
import java.io.Serializable;

import javax.swing.JFrame;

import org.LiveGraph.demoDataSource.LiveGraphDemo;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.DefaultXYDataset;
import org.jfree.ui.RectangleEdge;
import org.jfree.util.ShapeUtilities;


public class Synapse implements Serializable{
	private static final long serialVersionUID = 1L;
	protected String synapseType;
	protected Neuron preNode;
	protected Neuron postNode;
	protected boolean preNodeFired;
	protected boolean postNodeFired;
	protected double currentWeight;
	protected double STDPWeight;				
	protected double Aplus, Aminus;
	protected int delay;
	protected long spikePreQueue;
	protected long spikePostQueue;
	protected int startOfQueuePosition;
	protected int counter;
	protected int tetta;
	protected int tettaSyn;
	protected double g;
	protected double Asyn;
	protected double Esyn;
	protected Settings z;
	protected double currentSpikeRate; 
	protected double prevSpikeRate; 
	protected double currentPostNeuronSpikerate;
	protected double prevPostNeuronSpikerate;
	protected double deltaPreNeuronSpikeRate;
	protected double deltaPostNeuronSpikeRate;
	protected long spikeHistory;
	//Skinner Hume Pavlov stuff
	protected double alfa;  
	protected double betta; 
	protected double prevTracePreSpikeRate; //delta y //se side 11 Hokland 2011
	protected double prevTracePostSpikeRate; //delta x

	protected double learningConstant;
	protected int synapseTypeInt; //Exitatory=1, Inhibitory=-1, No sign restriction = 0
	protected double STDPLearningConstant;
	protected int lastFiredTimeSyn;
	protected int spikeDistanceSyn;
	protected double lastTwoSpikeRateConstantSyn;
	protected int spikeRateType;
	protected double spikeRateSyn;				
	protected double spikeRateAlfa;
	protected int durationOfPlot;
	protected MyLiveGraph liveGraph;

	public Synapse(Neuron preNode, Neuron postNode, int synType, double Asyn, Settings z, double initWeight){	//LearningRule learningRule
		durationOfPlot= Math.min(z.getTimesteps()-50-z.getWeightUpdateRate(), 100000)/z.getWeightUpdateRate();
		this.preNode=preNode;
		this.postNode=postNode;
		this.Asyn = Asyn;
		this.z = z;
		tettaSyn = z.getTettaSyn();
		learningConstant = z.getLearningConstant();
		this.Esyn = 0.0;
		this.g = 0;
		STDPLearningConstant = z.getSTDPLearningConstant();
		synapseType = "ST";
		synapseTypeInt = synType;
		alfa= z.getAlfa();
		betta = z.getBetta();
		prevTracePostSpikeRate = 0;
		prevTracePreSpikeRate = 0;
		// Default values before updating delays.
		this.delay = 	1;
		startOfQueuePosition = 1;
		this.currentWeight =  initWeight;		//
		STDPWeight = 0;

		preNodeFired = false;
		postNodeFired = false;
		lastFiredTimeSyn=0;
		spikeDistanceSyn=1000;
		lastTwoSpikeRateConstantSyn = z.getLastTwoSpikerateConstant();
		spikeRateType= z.getSpikeRateType();	//0 = trace, 1= distance , 2 = sigmoid window
		spikeRateSyn=0;
		spikeRateAlfa= z.getSpikeRateAlfa();
		spikePreQueue = spikePostQueue = spikeHistory= 0;
		//STDP STUFF
		tetta = z.getDefaultTetta();
		Aminus = 0.12;
		Aplus = 0.1;
	}

	public void updateSpikes() {
		if (counter!=0) counter--;				// Decrement if not yet zero.
		spikeRateSyn= spikeRateAlfa*spikeRateSyn ;

		if(delay==0) {
			spikeHistory=spikeHistory << 1;
			if(preNodeFired) {
				spikeRateSyn+= (1-spikeRateAlfa);
				spikeHistory+=1;
				g+=1.0;
				preNodeFired = false;
				//IF Spikerate is calculated based only on the last two spikes
				spikeDistanceSyn = SpikingNetwork.currentTime-lastFiredTimeSyn+delay;
				lastFiredTimeSyn = SpikingNetwork.currentTime;
				//--------------------------------------------------------	
			}
		}else{
			spikePostQueue = spikePostQueue << 1; // Starting off by shifting the post queue by one, adding a zero to the end of it.
			spikeHistory= spikeHistory << 1 ;

			if ((spikePreQueue & 1L)!= 0)  {		// Is there a spike coming next? (In the least significant bit position)
				g+=1.0;								// If there is, increase g
				spikePostQueue += 1;					// And then change the value of the end of the postQueue to 1.
				spikeHistory += 1;
				if(counter!=0) {			//Spike while cooldown is still active, penalize weight.
					updateWeightPre();
				}
			}

			spikePreQueue = spikePreQueue >>> 1;
			if(preNodeFired) { 
				spikeRateSyn+= (1-spikeRateAlfa);
				//IF Spikerate is calculated based only on the last two spikes
				spikeDistanceSyn = SpikingNetwork.currentTime-lastFiredTimeSyn+delay;
				lastFiredTimeSyn = SpikingNetwork.currentTime;
				//--------------------------------------------------------
				spikePreQueue += startOfQueuePosition;	//setter inn en spike i posisjon "dealy" fra slutten av kOen
				preNodeFired = false;

			}
			// Lring skjer nr den postsynaptiske noden fyrer.
			if(postNode.hasFired) {// postNodeFired
				updateWeightPost(); //STDP
				spikePostQueue = 0; // Reset the queue, learning is done.
				counter = tetta;
			}
		}	
		updatePsc();
	}

	//STDP UPDATE RULES
	public void updateWeightPre() {			
		if(STDPLearningConstant>0){
			STDPWeight -= Aminus*Math.exp((counter/tetta))*STDPLearningConstant;
			currentWeight+=STDPWeight;
		}
	};

	public void updateWeightPost() {
		if(STDPLearningConstant>0){										//IF STDP activated
			for(int i=0; i< Long.SIZE; i++) {
				if ((spikePostQueue & (1L << i)) != 0)
				{
					STDPWeight += Aplus*Math.exp(-i/tetta);
				}
			}
			STDPWeight*=STDPLearningConstant;
			currentWeight+=STDPWeight;
			if(liveGraph!=null){

				liveGraph.updateDataSets(new double[]{SpikingNetwork.currentTime,preNode.getSpikerate(delay), postNode.getSpikerate(delay),currentWeight,STDPWeight, 0});				
			}
		}
	}
	//------------------------

	public void updatePsc(){
		g = g -(g/tettaSyn);
		postNode.increaseInput(Asyn * currentWeight * g , preNode.name);	//(Esyn - postNode.getV())
	}
	public boolean isPreNodeFired() {
		return preNodeFired;
	}
	public void setPreNodeFired() {
		this.preNodeFired = true;
	}
	public void setPreNodeFired(boolean fired) {
		this.preNodeFired = fired;
	}
	public boolean isPostNodeFired() {
		return postNodeFired;
	}
	public void setPostNodeFired() {
		this.postNodeFired = true;
	}
	public void setPostNodeFired(boolean fired) {
		this.postNodeFired = fired;
	}
	public double getCurrentWeight() {
		return currentWeight;
	}
	public void setWeight(double weight) {
		this.currentWeight = weight;	
	}
	public Neuron getPreNode() {
		return preNode;
	}
	public Neuron getPostNode() {
		return postNode;
	}
	public int getDelay() {
		return delay;
	}
	public void setDelay(int delay) {
		this.delay = delay;
		startOfQueuePosition = (int)Math.pow(2, this.delay-1);
	}
	public void updateWeight() {
		// Learning rules here - specific for Hume, Pavlov and Skinner synapses;
	}
	public String getType() {
		return synapseType;
	}
	public void setAsyn(double asyn) {
		Asyn=asyn;
	}
	public double getAsyn() {
		return Asyn;
	}

	protected void showGraph(double[] time, double[] weight, double[] trace, String tracePrePost, double[] preS, double[] postS, double[] deltaPreSR, String deltaPrePost, String from,String to, String type){
		DefaultXYDataset dataWeight = new DefaultXYDataset();
		DefaultXYDataset dataTrace = new DefaultXYDataset();
		XYLineAndShapeRenderer rendForce=new XYLineAndShapeRenderer(true, false);
		XYLineAndShapeRenderer rendForce2=new XYLineAndShapeRenderer(true, false);

		dataWeight.addSeries("preSpikeRate",new double[][]{time,preS});
		rendForce.setSeriesPaint(dataWeight.getSeriesCount()-1, Color.RED);
		rendForce.setSeriesStroke(dataWeight.getSeriesCount()-1, new BasicStroke(1.5f));

		dataWeight.addSeries("postSpikerate",new double[][]{time,postS});
		rendForce.setSeriesPaint(dataWeight.getSeriesCount()-1, Color.BLUE);
		rendForce.setSeriesStroke(dataWeight.getSeriesCount()-1, new BasicStroke(1.5f));

		dataWeight.addSeries("Weight", new double[][]{time,weight});
		rendForce.setSeriesPaint(dataWeight.getSeriesCount()-1, Color.GREEN);
		rendForce.setSeriesStroke(dataWeight.getSeriesCount()-1, new BasicStroke(1.5f));

		if(tracePrePost.equals("tracePreNode")){	
			dataTrace.addSeries(tracePrePost,new double[][]{time,trace});
			rendForce2.setSeriesPaint(dataTrace.getSeriesCount()-1, Color.RED);
			rendForce2.setSeriesStroke(dataTrace.getSeriesCount()-1, new BasicStroke(1.5f));

			dataTrace.addSeries(deltaPrePost,new double[][]{time,deltaPreSR});
			rendForce2.setSeriesPaint(dataTrace.getSeriesCount()-1, Color.BLUE);
			rendForce2.setSeriesStroke(dataTrace.getSeriesCount()-1, new BasicStroke(1.5f));
		}else{
			dataTrace.addSeries(tracePrePost,new double[][]{time,trace});
			rendForce2.setSeriesPaint(dataTrace.getSeriesCount()-1, Color.BLUE);
			rendForce2.setSeriesStroke(dataTrace.getSeriesCount()-1, new BasicStroke(1.5f));

			dataTrace.addSeries(deltaPrePost,new double[][]{time,deltaPreSR});
			rendForce2.setSeriesPaint(dataTrace.getSeriesCount()-1, Color.RED);
			rendForce2.setSeriesStroke(dataTrace.getSeriesCount()-1, new BasicStroke(1.5f));
		}

		JFreeChart weightChart = ChartFactory.createXYLineChart(
				type+ "   "+from+" -> "+ to,"Time ms","Weight, spikerate",
				dataWeight,PlotOrientation.VERTICAL,true, true, false);
		weightChart.getPlot().setBackgroundPaint(Color.WHITE);//new Color(0xBFEFFF));
		weightChart.setBackgroundPaint(Color.WHITE);

		JFreeChart traceChart = ChartFactory.createXYLineChart(
				type+ "   "+from+" -> "+ to,"Time ms","Trace spikerate",
				dataTrace,PlotOrientation.VERTICAL,true, true, false);
		traceChart.getPlot().setBackgroundPaint(Color.WHITE);//new Color(0xBFEFFF));
		traceChart.setBackgroundPaint(Color.WHITE);

		rendForce.setSeriesShape(0, ShapeUtilities.createDiamond(1));
		rendForce2.setSeriesShape(0, ShapeUtilities.createDiamond(1));

		weightChart.getXYPlot().setRenderer(rendForce);
		weightChart.getXYPlot().setDomainGridlinesVisible(true);  
		weightChart.getXYPlot().setRangeGridlinesVisible(true);  
		weightChart.getXYPlot().setRangeGridlinePaint(Color.black);  
		weightChart.getXYPlot().setDomainGridlinePaint(Color.black);
		weightChart.getLegend().setPosition(RectangleEdge.BOTTOM);

		traceChart.getXYPlot().setRenderer(rendForce2);
		traceChart.getXYPlot().setDomainGridlinesVisible(true);  
		traceChart.getXYPlot().setRangeGridlinesVisible(true);  
		traceChart.getXYPlot().setRangeGridlinePaint(Color.black);  
		traceChart.getXYPlot().setDomainGridlinePaint(Color.black);
		traceChart.getLegend().setPosition(RectangleEdge.BOTTOM);

		JFrame forceframe = new JFrame(type);
		forceframe.setContentPane(new ChartPanel(weightChart));
		forceframe.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		forceframe.setLocation(772, 0);
		forceframe.pack();
		forceframe.setVisible(true);

		JFrame forceframe2 = new JFrame(type);
		forceframe2.setContentPane(new ChartPanel(traceChart));
		forceframe2.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		forceframe2.setLocation(772, 400);
		forceframe2.pack();
		forceframe2.setVisible(true);
	}
}
